import { default as default_2 } from 'react';
import { immutableNoteCallerNodeName } from 'shared-react/nodes/scripture/usj/ImmutableNoteCallerNode';
import { LoggerBasic } from 'shared-react/plugins/logger-basic.model';
import { ScriptureReference } from 'platform-bible-react';
import { Usj } from 'shared/converters/usj/usj.model';
import { UsjNodeOptions } from 'shared-react/nodes/scripture/usj/usj-node-options.model';
import { usjToUsxString } from 'shared/converters/usj/usj-to-usx';
import { usxStringToUsj } from 'shared/converters/usj/usx-to-usj';

export declare const DEFAULT_VIEW_MODE = "formatted";

/**
 * Scripture Editor for USJ. Created for use in Platform.Bible.
 * @see https://github.com/usfm-bible/tcdocs/blob/usj/grammar/usj.js
 *
 * @param props.defaultUsj - Default USJ Scripture data.
 * @param props.ref - Forward reference for the editor.
 * @param props.scrRef - Scripture reference that controls the cursor in the Scripture.
 * @param props.setScrRef - Scripture reference set callback function when the reference changes in
 *   the editor as the cursor moves.
 * @param props.options - Options to configure the editor.
 * @param props.onChange - Callback function when USJ Scripture data has changed.
 * @param props.logger - Logger instance.
 * @returns the editor element.
 */
export declare const Editor: default_2.ForwardRefExoticComponent<EditorProps<LoggerBasic> & default_2.RefAttributes<EditorRef>>;

/** Options to configure the editor. */
export declare type EditorOptions = {
    /** Is the editor readonly or editable. */
    isReadonly?: boolean;
    /** Is the editor enabled for spell checking. */
    hasSpellCheck?: boolean;
    /** View options. */
    view?: ViewOptions;
    /** Options for each editor node:
     * @param nodes[].noteCallers - Possible note callers to use when caller is '+' for
     *   ImmutableNoteCallerNode.
     * @param nodes[].onClick - Click handler method for ImmutableNoteCallerNode.
     */
    nodes?: UsjNodeOptions;
};

declare type EditorProps<TLogger extends LoggerBasic> = {
    /** Initial Scripture data in USJ form. */
    defaultUsj?: Usj;
    /** Scripture reference. */
    scrRef?: ScriptureReference;
    /** Set Scripture reference callback function. */
    setScrRef?: (scrRef: ScriptureReference) => void;
    /** Options to configure the editor. */
    options?: EditorOptions;
    /** Callback function when USJ Scripture data has changed. */
    onChange?: (usj: Usj) => void;
    /** Logger instance. */
    logger?: TLogger;
};

/** Forward reference for the editor. */
export declare type EditorRef = {
    /** Method to focus the editor. */
    focus(): void;
    /** Method to set the USJ Scripture data. */
    setUsj(usj: Usj): void;
};

/**
 * Get view option properties based on the view mode.
 * @param viewMode - View mode of the editor.
 * @returns the view options if the view exists, the default options if the viewMode is undefined,
 *   `undefined` otherwise.
 */
export declare function getViewOptions(viewMode?: string | undefined): ViewOptions | undefined;

export { immutableNoteCallerNodeName }

export { Usj }

export { UsjNodeOptions }

export { usjToUsxString }

export { usxStringToUsj }

export declare type ViewMode = ViewNameKey;

declare const viewModeToViewNames: {
    formatted: string;
    unformatted: string;
};

declare type ViewNameKey = keyof typeof viewModeToViewNames;

export declare type ViewOptions = {
    /** USFM markers are visible, editable or hidden */
    markerMode: "visible" | "editable" | "hidden";
    /** does the text have spacing including indenting */
    hasSpacing: boolean;
    /** is the text in a formatted font */
    isFormattedFont: boolean;
};

/**
 * Convert view options to view mode if the view exists.
 * @param viewOptions - View options of the editor.
 * @returns the view mode if the view is defined, `undefined` otherwise.
 */
export declare function viewOptionsToMode(viewOptions: ViewOptions | undefined): ViewMode | undefined;

export { }
